export function Attr(): void;
export function CDATASection(): void;
export function CharacterData(): void;
export function Comment(): void;
export function DocumentFragment(): void;
export function DocumentType(): void;
export function Element(): void;
export function Node(): void;
export function ShadowRoot(): void;
export function Text(): void;
export function SVGElement(): void;
export function illegalConstructor(): never;
export namespace Facades {
    export { Attr };
    export { CDATASection };
    export { CharacterData };
    export { Comment };
    export { DocumentFragment };
    export { DocumentType };
    export { Element };
    export { Node };
    export { ShadowRoot };
    export { Text };
    export { SVGElement };
}
